(**************************************************************************)
(*                              insertion.v                               *)
(*                                                                        *)
(*             Fonction annote insertion et preuve de correction         *)
(*                                                                        *)
(* Nicolas Magaud July 1998                                               *)
(**************************************************************************)

Require Programs.
Require Export lemmas.

Parameter N:Z.
Global Variable A : array N of Z.   (* the array *)

Correctness insertion
   fun (n:Z) ->
    {((sorted_array A `0` `n-1`) /\ `1<=n<N`)}
    let v = ref A[n] in
    let j = ref n in
    let fini = ref false in
    begin
      while ((!j > 0) and (not !fini)) do        
        { invariant ((  (global_inv N A j n v fini)
                        /\ (inv_0 N A j n v)
                        /\ (inv_n N A j n v fini)
                     )
                    /\ (permut (store A j v) A@0)
                    /\ (`0<=j<=n`)
                    ) as I
          variant (fini,j) for R }
        if (A[(Zpred !j)] > !v)
           then begin
                  A[!j] := A[(Zpred !j)];
                  j := (Zpred !j)
                end
           else begin fini := true end
      done;               
      A[!j]:=!v 
    end
  { (sorted_array A `0` n) /\ (permut A A@)}.

Proof.

(* 1 : ordre lexicographique bien fonde *)
Exact R_wf.

(* ecriture dans la case `j-1` du tableau - test du if ? *)
Abstract (Unfold Zpred; Omega).

Split.
(* 2 : conservation de l'invariant avec test du if a true 
       et dcroissance du variant *)

(* dcroissance du variant *)
Abstract 
  (Unfold R ; Right ; Split ; [Trivial with v62 | Unfold Zwf ; Unfold Zpred ; Omega]).

(* conservation de l'invariant *)
Decompose [and] I ; Clear I.
Split. (* inv - permut *)

Cut (`j0=n`)\/(`0<j0<n`).
Intro Cut_1; Elim Cut_1 ; Clear Cut_1.

(* j0=n *)
  Intro j_eq_n.
  Elim H3 ; Try Assumption ; Intros.
  (* Decompose [and] H6. *)
  Cut (`n=1`)\/(`n>1`).
  Intro Cut_2; Elim Cut_2 ; Clear Cut_2.
    (* n=1 *)
    Intro n_eq_1.
    Split.
     Abstract (Unfold global_inv ; Intros ; 
               Absurd `0 < (Zpred j0) < n` ; 
      	       [Unfold Zpred; Omega | Assumption]).
    Split.
       (* preuve *)
        Unfold inv_0 ; Intro.
    
	Split.
	Rewrite H7.
	Rewrite j_eq_n.
	Rewrite n_eq_1.
	Apply two_elements_array_with_same_value_is_sorted with i:=`0` ; 
      	  Try (Abstract Omega).
	Rewrite store_def_2 ; Try (Abstract Omega).
	Abstract (Rewrite store_def_1 ; Omega).

	Abstract (Rewrite store_def_2 ; Omega).
 
    Abstract (Unfold inv_n ; Intros ; Rewrite j_eq_n in H7 ;
              Absurd `(Zpred n) = n` ; [Unfold Zpred ; Omega | Assumption]).
    
    (* n>1 *)
    Intro.
    Split.
	Unfold global_inv ; Intros.
	Split.
	Apply right_reduction ; Try (Abstract Omega).
        Assumption.
	
	Split.
	Rewrite j_eq_n ; Unfold Zpred.
	Replace n with `n+(Zinv 1)+1` ; Try (Abstract Omega).
	Replace `n+(Zinv 1)+1+(-1)` with `n+(-1)` ; Try (Abstract Omega).
	Apply two_elements_array_with_same_value_is_sorted with i:=`n+(-1)` ;
      	  Try (Abstract Omega).
	Rewrite store_def_2 ; Try (Abstract Omega).
	Abstract (Rewrite store_def_1 ; Omega).
	Split.
	Unfold Zpred.
	Rewrite store_def_2 ; Try (Abstract Omega).
	Ring `j0+(-1)+1`.
	Rewrite store_def_1 ; Try (Abstract Omega).
	Unfold Zminus.
	Ring `j0+(-1)+(Zinv 1)`.
	Elim H3 ; Intros.
	Replace `-2+j0` with `j0+(-2)` ; Try (Abstract Omega).
	Replace `j0+(-1)` with `j0+(-2)+1` ; Try (Abstract Omega).
	Apply sorted_elements with n:=`0` m:=`j0-1` ; Try (Abstract Omega).
	Assumption.

    	Assumption.

	Split.
	Unfold Zpred.
	Rewrite store_def_2 ; Try (Abstract Omega).
	Apply Zlt_le_weak.
	Abstract (Unfold Zpred in Test1 ; Omega).
	Intro fini_eq_true.
	Decompose [and] Test2.
	Absurd fini0=false; 
      	  [Rewrite fini_eq_true ; Auto with v62 | Assumption].

    Split.
    Abstract (Unfold inv_0 ; Intro Zpred_j_eq_0;
              Unfold Zpred in Zpred_j_eq_0 ;
              Absurd `j0+(-1) = 0` ; Omega).
    
    Abstract (Unfold inv_n; Intro Zpred_j_eq_n;
              Unfold Zpred in Zpred_j_eq_n ; Absurd `j0+(-1) = n`; Omega).
    
    Abstract Omega. (* coupure : `n=1` \/ `n>1` *)

(* `0<jO<n` *)
  Intro.
  Elim H ; Try Assumption ; Intros.
  Decompose [and] H7 ; Clear H7.
  Cut (`j0=1`)\/(`j0>1`).
  Intro Cut_3 ; Elim Cut_3 ; Clear Cut_3.
  (* j0=1 *)
    Intro j_eq_1.
    Split.
    Abstract (Unfold global_inv ; Intros ;
              Absurd `j0 = 1` ; [ Unfold Zpred in H7 ; Omega | Omega ]).

	Split.
	Unfold inv_0; Intro.
	Split.
	Unfold Zpred.
	Apply left_extension_with_a_new_value ; Try (Abstract Omega).
	Assumption.
	Abstract (Rewrite store_def_2 ; Omega).

    Abstract (Unfold inv_n ; Intro ;
              Absurd `(Zpred j0) = n` ; [Unfold Zpred ; Omega | Omega]).

  (* j0 > 1 *)
    Split.
	Unfold global_inv; Intros.
	Split.
	Apply right_reduction ; Try (Abstract Omega).
        Assumption.
	Abstract (Unfold Zpred ; Omega).
	Split.
	Unfold Zpred ; Apply left_extension_with_a_new_value ; 
      	  Try (Abstract Omega).
	Assumption.
	Split.
	Unfold Zpred.
	Rewrite store_def_2 ; Try (Abstract Omega).
	Ring `j0+(-1)+1`.
	Rewrite store_def_1 ; Try (Abstract Omega).
	Apply sorted_elements with n:=`0` m:=`j0-1` ; Try (Abstract Omega).
	Assumption.
	Split.
	Rewrite store_def_2; Try (Abstract Omega).
	Abstract (Unfold Zpred ; Omega).

        Abstract (Intro fini_eq_true; Decompose [and] Test2 ;
                  Absurd fini0=false ; 
      	       	  [Rewrite fini_eq_true ; Auto with v62 | Assumption]).
    Split.
    Abstract (Unfold inv_0 ; Intro ;
              Absurd `(Zpred j0) = 0` ; [Unfold Zpred ; Omega | Assumption]).

    Abstract (Unfold inv_n ; Intro ;
              Absurd `(Zpred j0) = n` ; [Unfold Zpred ; Omega | Assumption]).
 (*les coupures *)    
    Abstract Omega.
    Abstract Omega.  (* `j0 = n`\/`0 < j0 < n` *)

(* permut *)
    Split.
    Apply permut_trans with t':=(store A0 j0 (A#[n])).
    Abstract (Unfold Zpred ; Apply shift_and_store_is_perm ; Omega).
    Assumption.

(* `0<=(Zpred j)<=n` *)    
    Abstract (Unfold Zpred ; Omega).


(* 3 : criture dans la case j du tableau ?*)
    Abstract Omega.


(* `0<j<N-1` *)    
    Abstract (Unfold Zpred ; Omega).

Split.

(* 4 : conservation de l'invariant avec test du if a false 
       et dcroissance du variant *)
(* dcroissance du variant *)
Decompose [and] Test2 ; Rewrite H1.
Abstract (Unfold R ; Left ; Red ; Auto with v62).

(* conservation de l'invariant *)
Decompose [and] I ; Clear I.
Split.
Cut (`j0=n`)\/(`0<j0<n`).
Intro Cut_4; Elim Cut_4 ; Clear Cut_4.
  (* j0=n *)
  Intro j_eq_n.
  Split.
  Abstract (Unfold global_inv ; Intros ;
            Absurd `j0=n` ; [Omega | Assumption]).
  Split.
  Abstract (Unfold inv_0 ; Intros ;
            Absurd `j0=0` ; [Omega | Assumption]).

      Unfold inv_n ; Intros.
      Elim H3 ; Try Assumption ; Intros.
      Decompose [and] H7 ; Clear H7.
      Split ; Try Assumption.
      Split ; Try Assumption.
      Intro true_eq_true ; Unfold Zpred in Test1 ; Unfold Zminus; Assumption.
 
  (* `0<j0<n` *)
  Intro.
  Split.
  Elim H ; Try Assumption ; Clear H ; Intros. 
  Decompose [and] H6 ; Clear H6.
  Unfold global_inv; Intros.
  Repeat Split ; Try Assumption.
  Intro true_eq_true ; Unfold Zpred in Test1 ; Unfold Zminus ; Assumption.
  Split.
  Abstract (Unfold inv_0 ; Intro ;
            Absurd `j0 = 0` ; [Omega | Assumption]).

  Abstract (Unfold inv_n ; Intro ;
            Absurd `j0 = n` ; [Omega | Assumption]).
  

  Abstract Omega. (* coupure *)
  
(* permut *)
Split.
Assumption.
(* `0 <= j0 <= n` *)
Abstract Omega.

(* 7 : entre de boucle *)
Decompose [and] Pre6 ; Clear Pre6.
Split.
Split.
  Abstract (Unfold global_inv ; Intros ;
            Absurd `0 < n < n` ; [Omega | Assumption]).
Split.
  Unfold inv_0 ; Intro.
  Split.
  Abstract (Apply single_element_array_is_sorted ; Omega).
  Abstract Omega.
  
  Unfold inv_n ; Intro.
  Split.
  Assumption.
  Split.
  Abstract Omega.
  Intro false_eq_true. Discriminate false_eq_true.
  Split.
  Apply eq_is_permut.
  Assumption.
  Abstract Omega.

(* 8 : sortie de boucle *)
Decompose [and] I ; Clear I.
Split.
Elim H1 ; Clear H1 ; Intro. 
(* `j0<=0` *)
Elim H ; Try Assumption ; Intros.
Replace `0` with j0 ; Try (Abstract Omega).
Apply left_substitution ; Try (Abstract Omega).
Assumption.
Abstract Omega.

(* fini0=true *)
Cut (`j0=0`)\/(`0<j0<n`)\/(`j0=n`).
Intro ; Decompose [or] H5 ; Intros.
    (* `j0=0` *)
    Rewrite H8.
    Elim H ; Try Assumption ; Intros.
    Apply left_substitution ; Try (Abstract Omega).
    Rewrite H8 in H7 ; Assumption.
    Rewrite H8 in H9 ; Assumption.
    (* `0<j0<n` *)
    Elim H0 ; Try Assumption ; Intros.
    Decompose [and] H9 ; Clear H9.
    Apply concat_of_two_sorted_arrays ; Try (Abstract Omega).
    Assumption.
    Assumption.
    Apply H14 ; Assumption.
    (* `j0=n` *)
    Rewrite H7.
    Elim H4 ; Try Assumption ; Intros.
    Decompose [and] H9 ; Clear H9.
    Replace n with `n-1+1` ; Try (Abstract Omega).
    Apply right_extension_with_a_new_value ; Try (Abstract Omega).
    Rewrite <- H7 ; Try Assumption.
    Rewrite H7 in H12.
    Unfold Zminus in H12 ; Unfold Zminus ; Ring `n+(Zinv 1)+1`.
    Apply H12 ; Assumption.

    Abstract Omega. (* la coupure *)

(* permut *)
Assumption.

(* 9 : dernire affectation dans la case j du tableau *)
Abstract Omega.

(* 10 : dfinition de la variable locale v *)
Abstract Omega.

Save.
