(**************************************************************************)
(*                             lemmas.v                                   *)
(*                                                                        *)
(*       Quelques lemmes spcifiques  la preuve du tri par insertion     *)
(*                                                                        *)
(* Nicolas Magaud July 1998                                               *)
(**************************************************************************)

Require Export Insertion_Specifications.
Require Export Permut.
Require Export ProgWf.
Require Export ZArithRing.
Require Export Omega.
     
(* un sous-tableau avec un seul lment est tri *)

Lemma single_element_array_is_sorted : (N:Z)(A:(array N Z))(n:Z)
      (`0<=n<N`)->(sorted_array A n n).
Proof.
Intros N A n ; Unfold sorted_array ; Intros.
Absurd `x<n` ; Assumption.
Save.

Hints Resolve single_element_array_is_sorted : datatypes.

(* un tableau tri de 0  j dans lequel on affecte une case d'indice suprieur
   ou gal  j est tri de 0  `j-1` *)

Lemma right_reduction : (N:Z)(A:(array N Z))(j:Z)(v:Z)(k:Z)
 `0<j<N` -> (sorted_array A `0` j)
  -> `j<=k<N` -> (sorted_array (store A k v) `0` `j-1`).
Proof.
Intros N A j v k H_j H_sorted H_k.
Apply no_effect ; Try Omega.

Apply sub_sorted_array with deb:=`0` fin:=j ; 
  [Assumption | Omega | Omega | Omega].

Save.

(* un tableau constitu d'lments tous identiques est tri *)

Lemma two_elements_array_with_same_value_is_sorted : 
  (N:Z)(A:(array N Z))(i:Z)
    `0<=i<N-1` ->  (A#[i] = A#[`i+1`]) -> (sorted_array A i `i+1`).
Proof.
Intros N A i H_i H_access.
Unfold sorted_array ; Intros.
Replace x with i ; Omega.

Save.

(* un tableau tri de i  j recevant A[i-1] en i reste tri si A[i-1] <= [i+1] *)

Lemma left_extension_with_a_new_value : 
 (N:Z)(A:(array N Z))(i:Z)(j:Z)
   `i>0`  -> `j<N`  -> (sorted_array A i j)
   -> (Zle A#[`i-1`]  A#[`i+1`]) 
   -> (sorted_array (store A i (A#[`i+-1`])) `i+-1` j).
Proof.
Intros N A i j H_i H_j H_sorted H_access.

Unfold sorted_array ; Intros.
Cut `x = i`\/`x <> i`.
Intro Cut ; Elim Cut ; Intros.
Unfold Zpred ; Rewrite H2.
Rewrite store_def_1 ; Try Omega.
Rewrite store_def_2 ; Try Omega.
Assumption.

Rewrite store_def_2 ; Try Omega.
Cut `i = x+1`\/`i <> x+1`.
Intro Cut2 ; Elim Cut2 ; Intros.
Rewrite H3.
Rewrite store_def_1 ; Try Omega.
Ring `x+1+-1` ; Apply Zle_n.

Rewrite store_def_2 ; Try Omega.
Apply H_sorted ; Omega.
Omega.

Omega.

Save.

(* fusion de 2 tableaux tris*)

Lemma concat_of_two_sorted_arrays : 
    (N:Z)(A:(array N Z))(n:Z)(i:Z)(v:Z)
    `0<n<N` -> `0<i<n`  -> (sorted_array A `0` `i-1`)
    -> (sorted_array A i n)
    -> (Zle v A#[i])
    -> (Zle A#[`i-1`] v) 
    -> (sorted_array (store A i v) `0` n).
Proof.
Intros N A n i v H_n H_i H_sorted1 H_sorted2 H_access1 H_access2.
(Unfold sorted_array; Intros).
Cut `x < i`\/`x = i`\/`x > i`.
Intro Cut ; Decompose [or] Cut ; Clear Cut.
(* `x < i` *)
Rewrite store_def_2 ; Try Omega.
Cut `x+1 < i`\/`x+1 = i`\/`x+1 > i`.
Intro Cut2 ; Decompose [or] Cut2 ; Clear Cut2.
    (* `x+1 < i` *)
    Rewrite store_def_2 ; Try Omega.
    Abstract (Apply H_sorted1; Omega).
    (* `x+1 = i` *)
    Rewrite H2.
    Rewrite store_def_1 ; Try Omega.
    Abstract (Replace x with `i-1` ; Omega).
    (* `x+1 > i` *)
    Abstract (Rewrite store_def_2; Omega).
Omega.
(* `x = i` *)
Rewrite H2.
Rewrite store_def_1 ; Try Omega.
Rewrite store_def_2 ; Try Omega.
Abstract (Apply Zle_trans with m:=(access A i) ; [Assumption | Apply H_sorted2; Omega]).
(* `x > i` *)
Rewrite store_def_2 ; Try Omega.
Cut `x+1 < i`\/`x+1 = i`\/`x+1 > i`.
Intro Cut3 ; Decompose [or] Cut3 ; Clear Cut3.
    (* `x+1 < i` *)
    Rewrite store_def_2 ; Omega.
    (* `x+1 = i` *)
    Rewrite H3.
    (Rewrite store_def_1; Try Omega).
    (* `x+1 > i` *)
    (Rewrite store_def_2 ; Try Omega).
    (Apply H_sorted2; Omega).

    Omega.

Omega.

Save.

(* extension  droite d'un tableau tri avec affectation  gauche *)

Lemma right_extension_with_a_new_value :
   (N:Z)(A:(array N Z))(i:Z)(j:Z)(v:Z)
   `i>=0` -> `j<N-1` -> `i<=j` -> (sorted_array A i j) 
   -> (Zle A#[j]  v) -> (sorted_array (store A `j+1` v) i `j+1`).
Proof.
Intros N A i j v H_i H_j H_ij H_sorted H_access.
(Apply right_extension; Try Omega).
(Apply no_effect; Try Omega).
Assumption.

(Rewrite store_def_2; Try Omega).
(Rewrite store_def_1; Try Omega).

Save.

(* un dcalage puis une affectation revient  une affectation 
 * directe dans la case adjacente  *)                                        

Lemma shift_and_store_is_perm : (N:Z)(A0:(array N Z))(j:Z)(v:Z)
   `0<j<N` -> 
   (permut (store (store A0 j A0#[`j+-1`]) `j+-1` v) (store A0 j v)).
Proof.
Intros.
Apply exchange_is_permut with i:=`j+-1` j:=j.
Apply exchange_c.
Omega.
Omega.
Rewrite store_def_1 ; Try Omega.
Rewrite store_def_1 ; Try Omega.
Rewrite store_def_2 ; Try Omega.
Rewrite store_def_1 ; Try Omega.
Rewrite store_def_2 ; Omega.

Intros.
Rewrite store_def_2 ; Try Omega.
Rewrite store_def_2 ; Try Omega.
Rewrite store_def_2 ; Omega.
Save.

(* 2 tableaux gaux sont une permutation l'un de l'autre *)

Lemma eq_is_permut : (N:Z)(A0:(array N Z))(j:Z)
   `0<=j<N` -> 
   (permut (store A0 j A0#[j])  A0).
Proof.
Intros.
Apply exchange_is_permut with i:=j j:=j.
Unfold exchange.
Constructor.
Assumption. Assumption.
Rewrite store_def_1. Reflexivity. Assumption.
Rewrite store_def_1. Reflexivity. Assumption.
Intros. Rewrite store_def_2. Reflexivity. Assumption. Assumption. Omega.
Save.

(* l'ordre lexicographique form de l'ordre sur les boolens (Zbool) et
 * celui sur les entiers relatifs positifs (Zwf `0`) 
 * est un ordre bien fond *)

Theorem R_wf : (well_founded bool*Z R).
Proof.
Unfold well_founded.
Induction a.
Intros b z.
Generalize (Zwf_well_founded `0`).
Intro H.
Unfold well_founded in H.
Generalize b.
Elim (H z).
Intros x H0 H1 b0.
Case b0.
Apply Acc_intro.
Induction y.
Intros b' z'.
Simpl.
Intro H2.
Elim H2.

Case b'; Simpl; Contradiction. 
Intro H3.
Decompose [and] H3.
Rewrite H5.
Auto with v62.

Apply Acc_intro.
Induction y.
Intros b' z'.
Case b'.
Intro H2.
Elim (H z').
Intros.
Apply Acc_intro.
Induction y0.
Intros.
Red in H5.
Elim H5.
Case y1; Simpl; Contradiction. 

Intro.
Decompose [and] H6.
Rewrite H8.
Auto with v62.

Simpl.
Intro.
Elim H2.
Contradiction.

Intro H3.
Decompose [and] H3.
Auto with v62.
Save.

