(**************************************************************************)
(*                           tri_insertion.v                              *)
(*                                                                        *)
(*     Programme annot de tri par insertion et preuve de correction      *)
(*                                                                        *)
(* Nicolas Magaud July 1998                                               *)
(**************************************************************************)

Require Programs.
Require lemmas.

Parameter N:Z.
Global Variable A : array N of Z.  (* the array *)
Global Variable insertion : fun (n:Z)
                            returns u:unit
                            reads A
                            writes A
      	       	       	    pre (sorted_array A `0` `n-1`)/\(`1<=n<N`)
                            post ((sorted_array A `0` n)/\(permut A A@))
                            end.

Correctness tri
  {`N>0`}
  let i = ref 1 in
    begin
      while (!i < N) do
        { invariant ((sorted_array A `0` `i-1`)/\(permut A A@0)/\(`1<=i`)) as I
          variant `N-i` }
        (insertion !i) { (sorted_array A `0` i)/\(permut A A@) };
        i := (Zs !i)
      done
    end
    { (sorted_array A `0` `N-1`)/\(permut A A@0) }.

Proof.

(* 1 : prcondition de insertion *)
Split.
Elim I; Auto with v62.
Omega.
(* 2 : dcroissance du variant et conservation de l'invariant *)
(* dcroissance du variant *)
Split.
Unfold Zwf ; Omega.

(* conservation de l'invariant *)
Split.
Unfold Zs ; Unfold Zminus ; Ring `i0+1+(Zinv 1)`.
Elim Post3; Intros ; Assumption.
Split.
Decompose [and] I.
Decompose [and] Post3.
Apply permut_trans with t':=A0 ; [Assumption | Assumption ].
Omega.

(* 3 : entre de boucle *)
Split ; [ Unfold Zminus ; Ring `1+(-1)` ;
          Apply single_element_array_is_sorted; Omega 
        | Split ; [Apply permut_refl | Omega]].

(* 4 : sortie de boucle *)
Decompose [and] I.
Split.
Apply sub_sorted_array with deb:=`0` fin:=`i0-1` ;
  [Assumption | Omega | Omega |Omega ].
Assumption.

Save.



